@@ -100,7 +100,7 @@ int get_balloon(const char *mount_point, struct stat *st, int *outfd)
 
 	fd = open(mount_point, O_RDONLY);
 	if (fd < 0) {
-		ploop_err(errno, "Can't open mount_point");
+		ploop_err(errno, "Can't open mount point %s", mount_point);
 		return(SYSEXIT_OPEN);
 	}
 
@@ -108,7 +108,7 @@ int get_balloon(const char *mount_point, struct stat *st, int *outfd)
 	close(fd);
 
 	if (fd2 < 0) {
-		ploop_err(errno, "Can't ioctl mount_point");
+		ploop_err(errno, "Can't ioctl mount point %s", mount_point);
 		return(SYSEXIT_DEVIOC);
 	}
 
@@ -754,7 +754,7 @@ int ploop_balloon_check_and_repair(const char *device, const char *mount_point,
 
 	ret = freemap2freeblks(freemap, top_level, &freeblks, &n_free_blocks);
 	if (ret)
-		return ret;
+		goto err;
 	if (!repair) {
 		ploop_log(0, "Found %u free blocks. Consider using "
 		       "\"ploop-balloon repair\"", n_free_blocks);
@@ -766,7 +766,7 @@ int ploop_balloon_check_and_repair(const char *device, const char *mount_point,
 
 	ret = ioctl_device(fd, PLOOP_IOC_FREEBLKS, freeblks);
 	if (ret)
-		return ret;
+		goto err;
 	drop_state = 0;
 	freezed_a_h = freeblks->alloc_head;
 	if (freezed_a_h > reverse_map_len) {
@@ -787,7 +787,7 @@ int ploop_balloon_check_and_repair(const char *device, const char *mount_point,
 		goto err;
 	ret = ioctl_device(fd, PLOOP_IOC_RELOCBLKS, relocblks);
 	if (ret)
-		return ret;
+		goto err;
 
 	ploop_log(0, "TRUNCATED: %u cluster-blocks (%llu bytes)",
 			relocblks->alloc_head,
@@ -800,8 +800,11 @@ err:
 	}
 
 	// FIXME: close_delta()
-	close(balloonfd);
-	close(fd);
+	if (balloonfd >= 0)
+		close(balloonfd);
+	if (fd >= 0)
+		close(fd);
+	free(pfiemap);
 	free(freemap);
 	free(rangemap);
 	free(relocmap);
@@ -843,7 +846,7 @@ static int ploop_trim(const char *mount_point, __u64 minlen_b, __u64 cluster)
 
 	fd = open(mount_point, O_RDONLY);
 	if (fd < 0) {
-		ploop_err(errno, "Can't open mount_point");
+		ploop_err(errno, "Can't open mount point %s", mount_point);
 		return -1;
 	}
 
