@@ -53,13 +53,16 @@ int extend_delta_array(struct delta_array * p, char * path, int rw, int od_flags
 	struct delta * da;
 
 	da = realloc(p->delta_arr, (p->delta_max + 1)*sizeof(struct delta));
-	if (da == NULL)
-		return -1;
+	if (da == NULL) {
+		ploop_err(errno, "realloc");
+		return SYSEXIT_NOMEM;
+	}
 	p->delta_arr = da;
 
 	if (open_delta(&da[p->delta_max], path, rw, od_flags))
-		return -1;
+		return SYSEXIT_OPEN;
 	p->delta_max++;
+
 	return 0;
 }
 
@@ -130,7 +133,7 @@ int open_delta_simple(struct delta * delta, const char * path, int rw, int od_fl
 
 int open_delta(struct delta * delta, const char * path, int rw, int od_flags)
 {
-	struct ploop_pvd_header *vh;
+	struct ploop_pvd_header *vh = NULL;
 	void *p;
 	ssize_t res;
 	struct stat stat;
@@ -144,9 +147,9 @@ int open_delta(struct delta * delta, const char * path, int rw, int od_flags)
 
 	rc = delta->fops->fstat(delta->fd, &stat);
 	if (rc != 0) {
+		err = errno;
 		ploop_err(errno, "stat %s", path);
-		close_delta(delta);
-		return -1;
+		goto error;
 	}
 
 	delta->l1_dirty = 0;
@@ -154,64 +157,56 @@ int open_delta(struct delta * delta, const char * path, int rw, int od_flags)
 	delta->l2_cache = -1;
 	delta->dirtied = 0;
 
-	if (posix_memalign(&p, 4096, SECTOR_SIZE)) {
-		close_delta(delta);
-		return -1;
-	}
+	if ((err = p_memalign(&p, 4096, SECTOR_SIZE)))
+		goto error;
 	vh = p;
-
-	res = delta->fops->pread(delta->fd, vh, SECTOR_SIZE, 0);
+	delta->hdr0 = p;
+	/* read header */
+	res = delta->fops->pread(delta->fd, delta->hdr0, SECTOR_SIZE, 0);
 	if (res != SECTOR_SIZE) {
 		err = (res >= 0) ? EIO : errno;
 		ploop_err(errno, "read 1st sector of %s", path);
-		goto open_delta_failed;
-	}
-	delta->blocksize = vh->m_Sectors;
-	cluster = S2B(vh->m_Sectors);
-
-	if (posix_memalign(&p, 4096, cluster)) {
-		err = errno;
-		goto open_delta_failed;
-	}
-	delta->hdr0 = p;
-
-	if (posix_memalign(&p, 4096, cluster)) {
-		err = errno;
-		goto open_delta_failed;
-	}
-	delta->l2 = p;
-
-	res = delta->fops->pread(delta->fd, delta->hdr0, cluster, 0);
-	if (res != cluster) {
-		err = (res >= 0) ? EIO : errno;
-		ploop_err(errno, "read %s", path);
-		goto open_delta_failed;
+		goto error;
 	}
 
-	if (memcmp(vh->m_Sig, SIGNATURE_STRUCTURED_DISK, sizeof(vh->m_Sig)) ||
-	    vh->m_Type != PRL_IMAGE_COMPRESSED ||
-	    !is_valid_blocksize(vh->m_Sectors)) {
+	delta->version = ploop1_version(vh);
+	if (delta->version == PLOOP_FMT_ERROR) {
+		ploop_err(errno, "Unknown ploop image version in the header %s",
+				path);
+		err = EINVAL;
+		goto error;
+	}
+
+	if (vh->m_Type != PRL_IMAGE_COMPRESSED ||
+			!is_valid_blocksize(vh->m_Sectors))
+	{
 		ploop_err(errno, "Invalid image header %s", path);
 		err = EINVAL;
-		goto open_delta_failed;
+		goto error;
 	}
-	delta->alloc_head = stat.st_size / (vh->m_Sectors * SECTOR_SIZE);
 
+	delta->blocksize = vh->m_Sectors;
+	delta->alloc_head = stat.st_size / (vh->m_Sectors * SECTOR_SIZE);
 	delta->l1_size = vh->m_FirstBlockOffset / vh->m_Sectors;
-	delta->l2_size = vh->m_SizeInSectors / vh->m_Sectors;
+	delta->l2_size = get_SizeInSectors(vh) / vh->m_Sectors;
+
+	cluster = S2B(vh->m_Sectors);
+	if (p_memalign(&p, 4096, cluster)) {
+		err = errno;
+		goto error;
+	}
+	delta->l2 = p;
 
 	if (vh->m_DiskInUse && !(od_flags & OD_ALLOW_DIRTY)) {
 		ploop_err(0, "Image is in use %s", path);
 		err = EBUSY;
-		goto open_delta_failed;
+		goto error;
 	}
 
-	free(vh);
 	return 0;
 
-open_delta_failed:
+error:
 	close_delta(delta);
-	free(vh);
 	errno = err;
 	return -1;
 }
@@ -346,27 +341,29 @@ static int relocate_block(struct delta *delta, __u32 iblk, void *buf,
 			delta->l2_cache = l2_cluster;
 		}
 
-		if (delta->l2[l2_slot] == iblk * B2S(cluster))
+		if (delta->l2[l2_slot] == ploop_sec_to_ioff((off_t)iblk * delta->blocksize,
+					delta->blocksize, delta->version))
 			break;
 	}
 
 	if (clu >= delta->l2_size)
 		return 0; /* found nothing */
 
-	if (READ(delta, buf, cluster,
-				S2B(delta->l2[l2_slot]))) {
+	if (READ(delta, buf, cluster, S2B(ploop_ioff_to_sec(delta->l2[l2_slot],
+						delta->blocksize, delta->version)))) {
 		ploop_err(errno, "Can't read block to relocate");
 		return -1;
 	}
 
-	delta->l2[l2_slot] = delta->alloc_head++ * B2S(cluster);
+	delta->l2[l2_slot] = ploop_sec_to_ioff((off_t)delta->alloc_head++ * delta->blocksize,
+			delta->blocksize, delta->version);
 	if (delta->l2[l2_slot] == 0) {
 		ploop_err(0, "relocate_block: delta->l2[l2_slot] == 0");
 		return -1;
 	}
 
-	if (WRITE(delta, buf, cluster,
-				S2B(delta->l2[l2_slot]))) {
+	if (WRITE(delta, buf, cluster, S2B(ploop_ioff_to_sec(delta->l2[l2_slot],
+						delta->blocksize, delta->version)))) {
 		ploop_err(errno, "Can't write relocate block");
 		return -1;
 	}
@@ -402,18 +399,19 @@ int grow_delta(struct delta *odelta, off_t bdsize, void *buf,
 	int i, rc;
 	struct ploop_pvd_header vh;
 	struct ploop_pvd_header *ivh = &vh;
-	int i_l1_size, i_l2_size;
+	int i_l1_size;
 	int i_l1_size_sync_alloc = 0;
+	off_t i_l2_size;
 	int map_idx = 0;
 	__u64 cluster = S2B(odelta->blocksize);
 
 	assert(cluster);
 
 	memset(ivh, 0, sizeof(*ivh));
-	generate_pvd_header(ivh, bdsize, odelta->blocksize);
+	generate_pvd_header(ivh, bdsize, odelta->blocksize, odelta->version);
 	ivh->m_DiskInUse = SIGNATURE_DISK_IN_USE;
 	i_l1_size = ivh->m_FirstBlockOffset / ivh->m_Sectors;
-	i_l2_size = ivh->m_SizeInSectors / ivh->m_Sectors;
+	i_l2_size = get_SizeInSectors(ivh) / ivh->m_Sectors;
 
 	if (odelta->alloc_head < odelta->l1_size) {
 		ploop_err(0, "grow_delta: odelta->alloc_head < odelta->l1_size");
@@ -426,6 +424,10 @@ int grow_delta(struct delta *odelta, off_t bdsize, void *buf,
 		return -1;
 	}
 
+	/* Total number of image-blocks in the image file is lesser
+	 * than number of image-blocks for new index table. So,
+	 * we can simply nullify this gap, no relocation needed for this
+	 */
 	if (odelta->alloc_head < i_l1_size) {
 		i_l1_size_sync_alloc = i_l1_size - odelta->alloc_head;
 		memset(buf, 0, cluster);
@@ -515,19 +517,20 @@ int grow_raw_delta(const char *image, off_t append_size)
 	void *buf;
 	unsigned long i = 0;
 
-	if (posix_memalign(&buf, 4096, DEF_CLUSTER)) {
-		ploop_err(errno, "posix_memalign");
+	if (p_memalign(&buf, 4096, DEF_CLUSTER))
 		return SYSEXIT_MALLOC;
-	}
+
 	memset(buf, 0, DEF_CLUSTER);
 
-	if (open_delta_simple(&delta, image, O_WRONLY, OD_NOFLAGS))
-		return SYSEXIT_OPEN;
+	if (open_delta_simple(&delta, image, O_WRONLY, OD_NOFLAGS)) {
+		ret = SYSEXIT_OPEN;
+		goto err1;
+	}
 
 	if(delta.fops->fstat(delta.fd, &stat)) {
 		ploop_err(errno, "fstat");
-		close_delta(&delta);
-		return SYSEXIT_READ;
+		ret = SYSEXIT_READ;
+		goto err;
 	}
 
 	pos = stat.st_size;
@@ -548,6 +551,7 @@ int grow_raw_delta(const char *image, off_t append_size)
 
 	if (delta.fops->fsync(delta.fd)) {
 		ploop_err(errno, "fsync");
+		ret = SYSEXIT_FSYNC;
 		goto err;
 	}
 	ret = 0;
@@ -557,5 +561,97 @@ int grow_raw_delta(const char *image, off_t append_size)
 
 err:
 	close_delta(&delta);
+err1:
+	free(buf);
+
+	return ret;
+}
+
+int ploop_grow_raw_delta_offline(const char *image, off_t new_size)
+{
+	int ret;
+	off_t old_size;
+
+	ret = read_size_from_image(image, 1, &old_size);
+	if (ret)
+		return ret;
+
+	new_size = (new_size + (4096 >> 9) - 1) & ~((4096 >> 9) - 1);
+
+	if (new_size == old_size)
+		return 0;
+
+	if (new_size < old_size) {
+		/* Use truncate(1) for offline truncate of raw delta */
+		ploop_err(0, "Error: new size %llu is less than "
+				"the old size %llu",
+				(unsigned long long)new_size,
+				(unsigned long long)old_size);
+		return SYSEXIT_PARAM;
+	}
+
+	return grow_raw_delta(image, (new_size - old_size) << 9);
+}
+
+int ploop_grow_delta_offline(const char *image, off_t new_size)
+{
+	off_t old_size;
+	struct ploop_pvd_header *vh;
+	struct ploop_pvd_header new_vh;
+	struct delta delta = {};
+	void *buf = NULL;
+	int ret = 0;
+
+	if (open_delta(&delta, image, new_size ? O_RDWR : O_RDONLY, OD_OFFLINE))
+		return SYSEXIT_OPEN;
+
+	vh = (struct ploop_pvd_header *)delta.hdr0;
+	old_size = get_SizeInSectors(vh);
+
+	generate_pvd_header(&new_vh, new_size, delta.blocksize, ploop1_version(vh));
+	if (get_SizeInSectors(&new_vh) == old_size)
+		goto out;
+
+	if (get_SizeInSectors(&new_vh) < old_size) {
+		ploop_err(0, "Error: new size %llu is less than "
+				"the old size %llu",
+				(unsigned long long)new_size,
+				(unsigned long long)old_size);
+		ret = SYSEXIT_PARAM;
+		goto out;
+	}
+
+	if (dirty_delta(&delta)) {
+		ploop_err(errno, "Failed to set dirty flag");
+		ret = SYSEXIT_WRITE;
+		goto out;
+	}
+
+	if (p_memalign(&buf, 4096, S2B(delta.blocksize))) {
+		ret = SYSEXIT_NOMEM;
+		goto out;
+	}
+
+	ret = grow_delta(&delta, get_SizeInSectors(&new_vh), buf, NULL);
+	if (ret)
+		goto out;
+
+	if (clear_delta(&delta)) {
+		ploop_err(errno, "Failed to clear dirty flag");
+		ret = SYSEXIT_WRITE;
+		goto out;
+	}
+
+	if (fsync(delta.fd)) {
+		ploop_err(errno, "fsync");
+		ret = SYSEXIT_FSYNC;
+		goto out;
+	}
+	ret = 0;
+
+out:
+	close_delta(&delta);
+	free(buf);
+
 	return ret;
 }
