@@ -19,10 +19,12 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <string.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/file.h>
+#include <sys/times.h>
 #include <fcntl.h>
 #include <limits.h>
 #include <signal.h>
@@ -37,6 +39,15 @@
 static int create_file(char *fname)
 {
 	int fd;
+	int ret;
+	char *dir;
+
+	dir = basename(fname); /* Note GNU basename is used here */
+	ret = mkdir(dir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+	if ((ret < 0) && (errno != EEXIST)) {
+		ploop_err(errno, "Can't create directory %s", dir);
+		return -1;
+	}
 
 	fd = open(fname, O_WRONLY | O_CREAT | O_TRUNC, 0600);
 	if (fd == -1) {
@@ -75,10 +86,23 @@ static int set_timer(timer_t *tid, unsigned int timeout)
 	return 0;
 }
 
+#define SEC_TO_NSEC(sec) ((clock_t)(sec) * 1000000000)
+static clock_t get_cpu_time(void)
+{
+	struct timespec ts;
+
+	if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
+		ploop_err(errno, "clock_gettime");
+		return (clock_t)-1;
+	}
+	return SEC_TO_NSEC(ts.tv_sec) + ts.tv_nsec;
+}
+
 static int do_lock(const char *fname, unsigned int timeout)
 {
 	int fd, r, _errno;
 	timer_t tid;
+	clock_t end = 0;
 	struct sigaction osa;
 	struct sigaction sa = {
 			.sa_handler = timer_handler,
@@ -89,16 +113,19 @@ static int do_lock(const char *fname, unsigned int timeout)
 		return -1;
 	}
 	if (timeout) {
-		/* all other signals should be set with SA_RESTART */
+		end = get_cpu_time();
+		if (end == (clock_t)-1)
+			goto err;
+		end += SEC_TO_NSEC(timeout);
 		sigaction(SIGRTMIN, &sa, &osa);
 		if (set_timer(&tid, timeout))
-			return -1;
+			goto err;
 	}
 	while ((r = flock(fd, LOCK_EX)) == -1) {
 		_errno = errno;
 		if (_errno != EINTR)
 			break;
-		if (timeout == 0)
+		if (timeout == 0 || get_cpu_time() < end)
 			continue;
 		_errno = EAGAIN;
 		break;
@@ -110,15 +137,17 @@ static int do_lock(const char *fname, unsigned int timeout)
 	if (r != 0) {
 		if (_errno == EAGAIN) {
 			ploop_err(_errno, "The %s is locked", fname);
-			close(fd);
-			return -1;
+			goto err;
 		} else {
 			ploop_err(_errno, "Error in flock(%s)", fname);
-			close(fd);
-			return -1;
+			goto err;
 		}
 	}
 	return fd;
+
+err:
+	close(fd);
+	return -1;
 }
 
 void ploop_unlock(int *lckfd)
