@@ -218,6 +218,29 @@ static void cleanup_kill_process(void *data)
 	kill(pid, SIGTERM);
 }
 
+static int ploop_execvp(const char *file, char *const argv[])
+{
+	char *const paths[] = {
+		"/sbin", "/bin",
+		"/usr/sbin", "/usr/bin",
+		"/usr/local/sbin", "/usr/bin",
+		NULL
+	};
+	int i, ret;
+
+	if (file[0] == '/')
+		return execv(file, argv);
+
+	for (i = 0; paths[i] != NULL; i++) {
+		char cmd[PATH_MAX];
+
+		snprintf(cmd, sizeof(cmd), "%s/%s", paths[i], file);
+		ret = execv(cmd, argv);
+	}
+
+	return ret;
+}
+
 int run_prg(char *const argv[])
 {
 	int pid, ret, status;
@@ -230,9 +253,19 @@ int run_prg(char *const argv[])
 	pid = fork();
 	if (pid == 0) {
 		int fd = open("/dev/null", O_RDONLY);
-		dup2(fd, STDIN_FILENO);
+		if (fd >= 0) {
+			dup2(fd, STDIN_FILENO);
+			close(fd);
+		}
+		else {
+			ploop_err(errno, "Can't open /dev/null");
+			return -1;
+		}
 
-		execvp(argv[0], argv);
+		ploop_execvp(argv[0], argv);
+
+		ploop_err(errno, "Can't exec %s", argv[0]);
+		return 127;
 	} else if (pid == -1) {
 		ploop_err(errno, "Can't fork");
 		return -1;
@@ -258,3 +291,14 @@ int run_prg(char *const argv[])
 
 	return -1;
 }
+
+int p_memalign(void **memptr, size_t alignment, size_t size)
+{
+	int ret;
+
+	ret = posix_memalign(memptr, alignment, size);
+	if (ret)
+		ploop_err(ret, "Memory allocation failed, posix_memalign");
+
+	return ret;
+}
